"""
Tests cancel-tcpip-forward
"""
from test_dropbear import *

import asyncssh
import asyncio
import socket
import errno
import random

async def run(addr, port):
    async with asyncssh.connect(addr, port = port) as conn:

        fwds = []
        COUNT=4

        def connect(host, port):
            print(f"connection from {host}:{port}")

        for x in range(COUNT):
            server = await conn.start_server(connect, '', 0,
                                     encoding='utf-8')
            print(f"opened fwd {server} port {server.get_port()}")
            fwds.append(server)

        # Check that they can be connected
        for f in fwds:
            c = socket.create_connection(("localhost", f.get_port()))

        # In case list ordering is important
        random.shuffle(fwds)

        for f in fwds:
            print(f"close fwd {f}")
            f.close()
            await f.wait_closed()

            # Check it's closed
            try:
                c = socket.create_connection(("localhost", f.get_port()))
                assert False, f"Expected {f} port {f.get_port()} to be closed"
            except OSError as e:
                assert e.errno == errno.ECONNREFUSED

def test_canceltcplisten(request, dropbear):
    opt = request.config.option
    host = opt.remote or LOCALADDR
    port = int(opt.port)

    asyncio.run(run(host, port))



from pathlib import Path
import sys

from test_dropbear import *

def test_reexec(request, dropbear):
	"""
	Tests that two consecutive connections have different address layouts.
	This indicates that re-exec makes ASLR work
	"""
	map_script = (Path(request.node.fspath).parent / "parent_dropbear_map.py").resolve()
	activate = own_venv_command()
	if activate == "":
		cmd = map_script
	else:
		# run within the same venv, for python deps
		cmd = f"{activate}; {map_script}"
	print(cmd)
	r = dbclient(request, cmd, capture_output=True, text=True)
	map1 = r.stdout.rstrip()
	print(r.stderr, file=sys.stderr)
	r.check_returncode()

	r = dbclient(request, cmd, capture_output=True, text=True)
	map2 = r.stdout.rstrip()
	print(r.stderr, file=sys.stderr)
	r.check_returncode()

	print(map1)
	print(map2)
	# expect something like
	# "563174d59000-563174d5d000 r--p 00000000 00:29 4242372                    /home/matt/src/dropbear/build/dropbear"
	assert map1.endswith('/dropbear') or map1.endswith('/dropbearmulti')
	a1 = map1.split()[0]
	a2 = map2.split()[0]
	print(a1)
	print(a2)
	# relocation addresses should differ
	assert a1 != a2




import subprocess
import tempfile

import pytest

keytypes = [
	"rsa", "rsa-4096",
	"ed25519",
	"ecdsa", "ecdsa-256", "ecdsa-384", "ecdsa-521",
	]

def parse_keytype(kt):
	if '-' in kt:
		return kt.split('-')
	else:
		return (kt, None)

@pytest.mark.parametrize("keytype", keytypes)
@pytest.mark.parametrize("keyformat", [None, "PEM"])
def test_from_openssh(request, tmp_path, keytype, keyformat):
	"""
	Convert OpenSSH to Dropbear format,
	PEM and OpenSSH internal
	"""
	opt = request.config.option
	kt, keybits = parse_keytype(keytype)

	if kt == 'dss' and keyformat is None:
		pytest.skip("dss doesn't support openssh format")

	os_kt = kt
	if os_kt == 'dss':
		# OpenSSH calls it 'dsa', Dropbear calls it 'dss'
		os_kt = 'dsa'

	os_key = tmp_path / 'oskey1'
	db_key = tmp_path / 'dbkey1'

	# Generate an OpenSSH key
	args = [
		opt.ssh_keygen,
		'-f', os_key,
		'-t', os_kt,
		'-N', '', # no password
	]
	if keybits is not None:
		args += ['-b', keybits]
	if keyformat:
		args += ['-m', keyformat]
	p = subprocess.run(args, check=True)

	# Convert to dropbear format
	args = [
		opt.dropbearconvert,
		'openssh', 'dropbear',
		os_key, db_key,
	]
	p = subprocess.run(args, check=True)

	# Compare pubkeys
	args = [
		opt.dropbearkey,
		'-f', db_key,
		'-y'
	]
	p = subprocess.run(args, check=True, stdout=subprocess.PIPE, text=True)
	db_pubkey = p.stdout.splitlines()[1].strip()
	os_pubkey = os_key.with_suffix('.pub').open().read().strip()
	# we compare the whole key including comment since it currently matches
	assert db_pubkey == os_pubkey

@pytest.mark.parametrize("keytype", keytypes)
def test_roundtrip(request, tmp_path, keytype):
	"""
	Dropbear's private key format is deterministic so
	we can compare round trip conversion. (OpenSSH's
	format has more variable comments and other fields).
	"""
	opt = request.config.option
	kt, keybits = parse_keytype(keytype)

	os_key = tmp_path / 'oskey1'
	db_key1 = tmp_path / 'dbkey1'
	db_key2 = tmp_path / 'dbkey2'

	# generate a key
	args = [
		opt.dropbearkey,
		'-t', kt,
		'-f', db_key1,
	]
	if keybits is not None:
		args += ['-s', keybits]
	p = subprocess.run(args, check=True)

	# convert to openssh
	args = [
		opt.dropbearconvert,
		'dropbear', 'openssh',
		db_key1, os_key,
	]
	p = subprocess.run(args, check=True)

	# Check ssh-keygen can read it
	args = [
		opt.ssh_keygen,
		'-f', os_key,
		'-y',
	]
	p = subprocess.run(args, check=True, text=True, stdout=subprocess.PIPE)
	os_pubkey = p.stdout.strip()

	# Compare public keys
	args = [
		opt.dropbearkey,
		'-f', db_key1,
		'-y',
	]
	p = subprocess.run(args, check=True, text=True, stdout=subprocess.PIPE)
	db_pubkey = p.stdout.splitlines()[1].strip()
	# comment may differ
	db_pubkey = db_pubkey.split(' ')[:2]
	os_pubkey = os_pubkey.split(' ')[:2]
	assert db_pubkey == os_pubkey

	# convert back to dropbear
	args = [
		opt.dropbearconvert,
		'openssh', 'dropbear',
		os_key, db_key2,
	]
	p = subprocess.run(args, check=True)
	# check the round trip is identical
	assert db_key1.open('rb').read() == db_key2.open('rb').read()



"""
Tests opening and closing several (up to 4) channels concurrently.
"""
from test_dropbear import *

import asyncssh
import asyncio
import random

async def run(addr, port):
    async with asyncssh.connect(addr, port = port) as conn:

        chans = []
        MAX=4

        for x in range(10000):
            if len(chans) < MAX:
                pipes = await conn.open_session(command = "df")
                chans.append(pipes)
                l = len(chans)
                print(f" add, len {l}")

            if random.random() < 0.2:
                i = random.randrange(0, len(chans))
                l = len(chans)
                print(f" del {i}/{l}")
                del chans[i]

def test_concurrent(request, dropbear):
    opt = request.config.option
    host = opt.remote or LOCALADDR
    port = int(opt.port)

    asyncio.run(run(host, port))



def pytest_addoption(parser):
    parser.addoption("--port", type=str, help="default is 2244 local, 22 remote")
    parser.addoption("--dbclient", type=str, default="../dbclient")
    parser.addoption("--dropbear", type=str, default="../dropbear")
    parser.addoption("--dropbearconvert", type=str, default="../dropbearconvert")
    parser.addoption("--dropbearkey", type=str, default="../dropbearkey")
    parser.addoption("--hostkey", type=str, help="required unless --remote")
    parser.addoption("--remote", type=str, help="remote host")
    parser.addoption("--user", type=str, help="optional username")
    parser.addoption("--ssh-keygen", type=str, default="ssh-keygen")

def pytest_configure(config):
    opt = config.option
    if not opt.hostkey and not opt.remote:
        raise Exception("--hostkey must be given")
    if not opt.port:
        if opt.remote:
            opt.port = "22"
        else:
            opt.port = "2244"




from test_dropbear import *
import signal
import queue
import socket
import os
from pathlib import Path

# Tests for server side authentication

# Requires keyfile and authorized_keys set up in github action build.yml
@pytest.mark.skipif('DBTEST_IN_ACTION' not in os.environ, reason="DBTEST_IN_ACTION not set")
def test_pubkeyinfo(request, dropbear):
	kf = str(Path.home() / ".ssh/id_dropbear_key2")
	r = dbclient(request, "-i", kf, "echo -n $SSH_PUBKEYINFO", capture_output=True)
	# stop at first space
	assert r.stdout.decode() == "key2"

@pytest.mark.skipif('DBTEST_IN_ACTION' not in os.environ, reason="DBTEST_IN_ACTION not set")
def test_pubkeyinfo_special(request, dropbear):
	kf = str(Path.home() / ".ssh/id_dropbear_key3")
	r = dbclient(request, "-i", kf, "echo -n $SSH_PUBKEYINFO", capture_output=True)
	# comment contains special characters so the SSH_PUBKEYINFO should not be set
	assert r.stdout.decode() == ""

@pytest.mark.skipif('DBTEST_IN_ACTION' not in os.environ, reason="DBTEST_IN_ACTION not set")
def test_pubkeyinfo_okchar(request, dropbear):
	kf = str(Path.home() / ".ssh/id_dropbear_key4")
	r = dbclient(request, "-i", kf, "echo -n $SSH_PUBKEYINFO", capture_output=True)
	# comment contains special characters so the SSH_PUBKEYINFO should not be set
	assert r.stdout.decode() == "key4,char"
